"""
Class describing a challenge
"""

import logging

import click
from git import Repo

from aicrowd.constants import ChallengeConstants
from aicrowd.errors import CONFIG_NOT_SET
from aicrowd.exceptions import CLIException


class Challenge:
    """
    CLI's interpretation of a challenge
    """

    def __init__(self):
        self.log: logging.Logger = logging.getLogger()
        self.info = None

    def load(self):
        """
        Check git config for CLI stuff
        """
        try:
            with Repo(".").config_reader() as git_conf:
                challenge_id = git_conf.get_value(
                    ChallengeConstants.CONFIG_SECTION_NAME,
                    ChallengeConstants.CONFIG_ID_KEY,
                )
                challenge_slug = git_conf.get_value(
                    ChallengeConstants.CONFIG_SECTION_NAME,
                    ChallengeConstants.CONFIG_SLUG_KEY,
                )

                self.info = {
                    ChallengeConstants.CONFIG_ID_KEY: challenge_id,
                    ChallengeConstants.CONFIG_SLUG_KEY: challenge_slug,
                }
                self.log.info("Read challenge config\n---\n%s\n---", self.info)
        except Exception as e:
            self.log.error("Error while reading the git config, %s", e)
            self.info = {}

    def get(self, key: str, ensure_exists: bool = False, help_msg: str = "") -> str:
        """
        Gets a key from config

        Prints help_msg if ensure_exists is True and key not found

        Args:
            key: the config setting you are looking for
            ensure_exists: should this die if key doesn't exist?
            help_msg: how to fix this issue

        Returns:
            the value for the key
        """
        if self.info is None:
            self.load()

        value = self.info.get(key)

        if value is None and ensure_exists:
            self.log.warning("Queried challenge config for %s but not found", key)
            raise CLIException(
                f"Challenge config property {key} not set", help_msg, CONFIG_NOT_SET
            )

        return value



"""
Submission subcommand
"""
import os
import sys

import click

from aicrowd.contexts import (
    pass_config,
    ConfigContext,
    pass_challenge,
    ChallengeContext,
)
from aicrowd.errors import UNKNOWN_ERROR, INVALID_FILE
from aicrowd.utils import AliasedGroup


@click.group(name="submission", cls=AliasedGroup)
def submission_command():
    """
    Create and view submissions
    """


@click.command(name="create")
@click.option(
    "-c",
    "--challenge",
    type=str,
    help="Specify challenge explicitly",
)
@click.option(
    "-f",
    "--file",
    "file_path",
    type=click.Path(),
    default="",
    help="The file to submit",
)
@click.option(
    "-d",
    "--description",
    type=str,
    help="Description",
    default="",
)
@click.option("--jupyter", is_flag=True, help="Bundle jupyter notebook")
@pass_challenge
@pass_config
def create_subcommand(
    config_ctx: ConfigContext,
    challenge_ctx: ChallengeContext,
    challenge: str,
    file_path: str,
    description: str,
    jupyter: bool,
):
    """
    Create a submission on AIcrowd
    """
    if not jupyter and not os.path.exists(file_path):
        click.echo(
            click.style(f"Submission file {file_path} does not exist", fg="red"),
            err=True,
        )
        sys.exit(INVALID_FILE)

    from aicrowd.submission import create
    from aicrowd.submission.exceptions import SubmissionException

    try:
        print(
            create(
                challenge,
                file_path,
                description,
                True,
                config_ctx,
                challenge_ctx,
                jupyter=jupyter,
            )
        )
    except SubmissionException as e:
        message = e.message
        # separated to avoid cases where message is none
        if not message:
            message = (
                "An unknown error occured. Please set verbosity to check full logs."
            )

        click.echo(click.style(message, fg="red"), err=True)
        if e.fix:
            click.echo(click.style(e.fix, fg="yellow"))

        sys.exit(e.exit_code)


@click.command(name="save-assets")
@click.option(
    "-c",
    "--challenge",
    type=str,
    help="Specify challenge explicitly",
)
def save_assets_subcommand(challenge: str):
    from aicrowd.submission.copy_assets import copy_assets

    copy_assets(challenge, colab_to_drive=True)


@click.command(name="load-assets")
@click.option(
    "-c",
    "--challenge",
    type=str,
    help="Specify challenge explicitly",
)
def load_assets_subcommand(challenge: str):
    from aicrowd.submission.copy_assets import copy_assets

    copy_assets(challenge, colab_to_drive=False)


submission_command.add_command(create_subcommand)
submission_command.add_command(save_assets_subcommand)
submission_command.add_command(load_assets_subcommand)



"""
For handling logins
"""

import sys
import logging

import click

from aicrowd.contexts import pass_config, ConfigContext


@click.command(name="login")
@click.option("--api-key", type=str, help="API Key from AIcrowd website")
@pass_config
def login_command(config_context: ConfigContext, api_key: str):
    """
    Log in using AIcrowd API Key
    """
    from aicrowd.auth import login
    from aicrowd.auth.exceptions import LoginException

    log = logging.getLogger()

    if api_key is None:
        log.info("API Key not provided in parameters, prompting")

        click.launch("https://www.aicrowd.com/participants/me")
        click.echo("Please copy paste the API Key")
        api_key = click.prompt("API Key")

    try:
        login(api_key, config_context)
    except LoginException as e:
        click.echo(click.style(e.message, fg="red"))
        if e.fix:
            click.echo(click.style(e.fix, fg="yellow"))
        sys.exit(e.exit_code)



import os
import re
from typing import Any, Dict

from aicrowd.notebook.helpers import read_notebook, write_notebook


class TargetBlocks:
    """
    Markdown cell heading in the notebook and their phase mappings
    """

    INSTALL_START = "^# Install packages"
    PRE_PROCESS_START = "^# Define preprocessing code|^# Import packages"
    TRAIN_START = "^# Training phase"
    PREDICT_START = "^# Prediction phase"
    SUBMIT = "^# Submit to AIcrowd"
    INSTALL_END = PRE_PROCESS_START
    TRAIN_END = PREDICT_START
    PRE_PROCESS_END = TRAIN_START


def split_notebook(file_path: str, output_dir: str):
    """
    Split the submission notebook into install.ipynb, train.ipynb and
    predict.ipynb

    **Note:** API key scrubbing should be done before this

    Args:
        file_path: Path to the submission notebook
        output_dir: Directory to place the output files in
    """
    print("Collecting notebook...")
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    write_installation_notebook(file_path, output_dir)
    write_training_notebook(file_path, output_dir)
    write_prediction_notebook(file_path, output_dir)


def is_not_empty_cell(cell: Dict[str, Any]) -> bool:
    """
    Check for non empty cells

    Args:
        cell: Jupyter cell

    Returns:
        True of the cell is not empty
    """
    return len(cell["source"]) > 0


def extraction_direction(status: bool, remove: bool) -> bool:
    if remove:
        return not status
    return status


def extract_partial_notebook(
    nb: Dict[str, Any], start_block: str, end_block: str, remove: bool = False
) -> Dict[str, Any]:
    """
    Collects or removes cells blocks withing start and end headers

    Args:
        nb: Input notebook
        start_block: Regular expression for the starting block
        end_block: Regular expression for the ending block
        remove: Determines if

    Examples:
        To extract the notebook cells within `start_block` and `end_block`
        >>>extract_partial_notebook(nb, start_block="^# Heading 1", end_block="^# Heading 2")

        To remove the notebook cells within similar start and end blocks, use
        `remove=True`

    Returns:
        Notebook with cells with/without between the start and end blocks
    """
    target_cells = []
    start_copy = False
    for _cell in nb["cells"]:
        if is_not_empty_cell(_cell):
            if start_copy or re.search(start_block, _cell["source"][0]):
                start_copy = True
            if re.search(end_block, _cell["source"][0]):
                if not remove:
                    break
                start_copy = False
            if extraction_direction(start_copy, remove):
                target_cells.append(_cell)
    nb["cells"] = target_cells
    return nb


def write_installation_notebook(file_path: str, output_dir: str):
    """
    Extracts and writes install.ipynb to the submission directory

    Args:
        file_path: Path to the submission notebook
        output_dir: Path to the submission directory
    """
    nb = read_notebook(file_path)
    partial_nb = extract_partial_notebook(
        nb, TargetBlocks.INSTALL_START, TargetBlocks.INSTALL_END
    )
    write_notebook(os.path.join(output_dir, "install.ipynb"), partial_nb)


def write_training_notebook(file_path: str, output_dir: str):
    """
    Extracts and writes train.ipynb to the submission directory

    Args:
        file_path: Path to the submission notebook
        output_dir: Path to the submission directory
    """
    nb = read_notebook(file_path)
    partial_nb = extract_partial_notebook(nb, ".*", TargetBlocks.TRAIN_END)
    partial_nb = extract_partial_notebook(
        partial_nb, TargetBlocks.INSTALL_START, TargetBlocks.INSTALL_END, remove=True
    )
    write_notebook(os.path.join(output_dir, "train.ipynb"), partial_nb)


def write_prediction_notebook(file_path: str, output_dir: str):
    """
    Extracts and writes predict.ipynb to the submission directory

    Args:
        file_path: Path to the submission notebook
        output_dir: Path to the submission directory
    """
    nb = read_notebook(file_path)
    partial_nb = extract_partial_notebook(nb, ".*", TargetBlocks.SUBMIT)
    partial_nb = extract_partial_notebook(
        partial_nb, TargetBlocks.INSTALL_START, TargetBlocks.INSTALL_END, remove=True
    )
    partial_nb = extract_partial_notebook(
        partial_nb, TargetBlocks.TRAIN_START, TargetBlocks.TRAIN_END, remove=True
    )
    write_notebook(os.path.join(output_dir, "predict.ipynb"), partial_nb)



import importlib
import json
import os
import re
import shutil
import subprocess
import time
from types import ModuleType
from typing import List, Dict, Any, Tuple
from urllib.parse import unquote

import requests

from aicrowd.utils.jupyter import is_google_colab_env, mount_google_drive


def read_notebook(file_path: str) -> Dict[str, Any]:
    with open(file_path) as fp:
        nb = json.load(fp)
    return nb


def write_notebook(file_path: str, nb: Dict[str, Any]):
    with open(file_path, "w") as fp:
        json.dump(nb, fp)


def delete_expressions_from_notebook(
    expressions: List[str], file_path: str
) -> Dict[str, Any]:
    """
    Delete code lines from the notebook that match the regular expressions

    Args:
        expressions: List of regular expressions to match
        file_path: Path to the notebook
    """
    nb = read_notebook(file_path)

    # TODO: Duplicated code, refactor
    for _cell in nb["cells"]:
        # Match the lines only in code blocks
        if _cell["cell_type"] == "code":
            source_code = []
            for _code_line in _cell["source"]:
                matched = False
                for _expr in expressions:
                    if re.search(_expr, _code_line):
                        matched = True
                        break
                if not matched:
                    source_code.append(_code_line)
            _cell["source"] = source_code
    return nb


def convert_timestamp_to_epoch(timestamp: str) -> float:
    pattern = "%Y-%m-%dT%H:%M:%S.%fZ"
    return time.mktime(time.strptime(timestamp, pattern))


def get_default_jupyter_api_session_host() -> str:
    if is_google_colab_env():
        return "http://172.28.0.2:9000"
    return "http://127.0.0.1:8888"


def get_latest_jupyter_session() -> Dict[str, Any]:
    proxies = {"http": None, "https": None}
    (
        auto_detected_jupyter_host,
        auto_detected_jupyter_token,
    ) = get_jupyter_server_endpoint()
    jupyter_notebook_host = os.getenv("JUPYTER_NB_HOST", auto_detected_jupyter_host)
    response = requests.get(
        os.path.join(jupyter_notebook_host, "api/sessions"),
        proxies=proxies,
        headers={"Authorization": f"token {auto_detected_jupyter_token}"},
    )
    if not response.ok:
        raise requests.exceptions.HTTPError(
            f"Got invalid response from Jupyter: {response.text}"
        )
    sessions = response.json()
    latest_session = sorted(
        sessions,
        reverse=True,
        key=lambda x: convert_timestamp_to_epoch(
            x.get("kernel", {}).get("last_activity", -1)
        ),
    )[0]
    return latest_session


def get_notebook_path():
    if is_google_colab_env():
        return get_colab_notebook_path()
    else:
        return get_jupyter_notebook_path()


def get_jupyter_notebook_path():
    return os.path.join(os.getcwd(), get_latest_jupyter_session()["name"])


def get_colab_notebook_path():
    mount_path = mount_google_drive()
    return os.path.join(
        mount_path,
        "MyDrive/Colab Notebooks",
        unquote(get_latest_jupyter_session()["name"]),
    )


def bundle_notebook(submission_dir: str, notebook_name: str = None):
    if is_google_colab_env():
        raise NotImplementedError("Google colab submissions are not ready yet")
    else:
        bundle_original_jupyter_notebook(submission_dir, notebook_name)


def bundle_original_jupyter_notebook(submission_dir: str, notebook_name: str = None):
    submission_dir = submission_dir.replace(".zip", "")
    if notebook_name is None:
        notebook_name = get_jupyter_notebook_path()
    shutil.copy(notebook_name, os.path.join(submission_dir, "original_notebook.ipynb"))


def get_runtime_language() -> Tuple[str, str]:
    kernel = get_latest_jupyter_session()["kernel"]["name"]
    if kernel == "ir":
        return "r", kernel
    if kernel.startswith("python"):
        return "python", kernel


def write_aicrowd_config(submission_dir: str):
    language, kernel = get_runtime_language()
    config = {
        "language": language,
        "kernel": kernel,  # we do not use this value, it's added only for debugging
    }
    with open(os.path.join(submission_dir, "aicrowd.json"), "w") as fp:
        json.dump(config, fp)


def get_kernel_from_language(language: str) -> str:
    if language.lower() == "python":
        if is_google_colab_env():
            return "python3"
        return "python"
    if language.lower() == "r":
        return "ir"
    raise NotImplementedError("Unsupported language")


def import_module(module_name: str) -> ModuleType:
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError:
        pass


def get_jupyter_server_endpoint() -> Tuple[str, str]:
    """
    Returns the jupyter server endpoint along with the token

    Returns:
        Jupyter server endpoint and token
    """
    notebook_server = import_module("jupyter_server.serverapp")
    if not notebook_server:
        print("No jupyter lab module found. Using jupyter notebook.")
        notebook_server = import_module("notebook.notebookapp")

    if not notebook_server:
        print("No jupyter notebook module found. Using default values.")
        return get_default_jupyter_api_session_host(), ""

    session_list = list(notebook_server.list_running_servers())
    if len(session_list) > 1:
        print("WARNING: Got more than 1 jupyter server, selecting the latest session")
    if len(session_list) == 0:
        raise RuntimeError(
            "No jupyter server found. Did you start your jupyter server?"
        )
    return session_list[-1]["url"], session_list[-1]["token"]



import json
import os
import shutil
import subprocess

from aicrowd.utils.jupyter import execute_command, is_google_colab_env
from aicrowd.notebook.clean import clean_notebook
from aicrowd.notebook.helpers import (
    bundle_notebook,
    get_jupyter_notebook_path,
    write_aicrowd_config,
    get_kernel_from_language,
    get_notebook_path,
)
from aicrowd.notebook.split import split_notebook
from aicrowd.submission import create as submit_to_aicrowd

SUBMISSION_DIR = os.path.join(os.getcwd(), "submission")


def verify_submission():
    print("Validating the submission...")
    if not os.path.exists(SUBMISSION_DIR):
        raise FileNotFoundError("Please bundle the submission before local evaluation")
    with open(os.path.join(SUBMISSION_DIR, "aicrowd.json")) as fp:
        config = json.load(fp)
    nbconvert_cmd = "jupyter nbconvert --to notebook --ExecutePreprocessor.kernel_name={} --execute {}/{}"
    kernel = get_kernel_from_language(config["language"])
    for notebook in ["install.ipynb", "predict.ipynb"]:
        print(f"Executing {notebook}...")
        if execute_command(nbconvert_cmd.format(kernel, SUBMISSION_DIR, notebook)) != 0:
            raise RuntimeError(f"{notebook} failed to execute")


def bundle_submission(
    assets_dir: str,
    submission_zip_path: str = "submission.zip",
    notebook_name: str = None,
):
    # Make sure that assets dir is a direct child of current dir
    if os.path.basename(assets_dir) != assets_dir:
        raise NameError(
            "Assets directory should be a direct part of the current directory"
        )
    if not os.path.exists(assets_dir):
        print(f"WARNING: No assets directory at {assets_dir}... Creating one...")
        os.mkdir(assets_dir)
    if os.listdir(assets_dir) == 0:
        print("WARNING: Assets directory is empty")

    print(f"Using notebook: {get_jupyter_notebook_path()} for submission...")

    # Remove existing files from submission directoru
    if os.path.exists(SUBMISSION_DIR):
        print("Removing existing files from submission directory...")
        shutil.rmtree(SUBMISSION_DIR)
    os.mkdir(SUBMISSION_DIR)

    # Clean sensitive stuff from notebook
    cleaned_notebook_path = os.path.join(SUBMISSION_DIR, "original_notebook.ipynb")
    if notebook_name is None:
        notebook_name = get_notebook_path()
    clean_notebook(notebook_name, cleaned_notebook_path)

    # Breaks the notebook into {install, train, predict}.ipynb
    split_notebook(cleaned_notebook_path, SUBMISSION_DIR)
    # Write aicrowd.json
    write_aicrowd_config(SUBMISSION_DIR)
    # Copy the original notebook
    # bundle_notebook(submission_zip_path, notebook_name)
    # Copy assets
    shutil.copytree(assets_dir, os.path.join(SUBMISSION_DIR, assets_dir))
    # Zip everything
    shutil.make_archive(submission_zip_path.replace(".zip", ""), "zip", SUBMISSION_DIR)


def create_submission(
    challenge: str,
    description: str,
    assets_dir: str,
    submission_zip_path: str = "submission.zip",
    notebook_name: str = None,
    no_verify: bool = False,
    dry_run: bool = False,
):
    bundle_submission(assets_dir, submission_zip_path, notebook_name)
    if not no_verify:
        verify_submission()
    if not dry_run:
        submit_to_aicrowd(
            challenge=challenge,
            file_path=submission_zip_path,
            description=description,
            print_links=True,
        )



"""
Dataset related tasks
"""

import logging
import os
import fnmatch
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Iterable, List, Dict
from urllib.parse import urlparse
from urllib.request import Request, urlopen

import requests

# pylint: disable=unused-import
from aicrowd.challenge.helpers import must_get_api_key, parse_cli_challenge

# pylint: enable=unused-import
from aicrowd.constants import AICROWD_API_ENDPOINT, DATASETS_HOST
from aicrowd.utils import TqdmProgressBar
from aicrowd.utils.jupyter import is_google_colab_env


def get_datasets(challenge_id: int, api_key: str) -> List[dict]:
    """
    Queries AIcrowd API for datasets of this challenge

    Args:
        challenge_id: challenge id
        api_key: AIcrowd API Key

    Returns:
        Datasets for a particular challenge
    """
    log = logging.getLogger()

    r = requests.get(
        f"{AICROWD_API_ENDPOINT}/challenges/{challenge_id}/dataset",
        headers={"Authorization": f"Token {api_key}"},
    )

    if not r.ok:
        log.error("Request to API failed\nReason: %s\nMessage: %s", r.reason, r.text)
        return [{}]

    try:
        return r.json()
    except Exception as e:
        log.error("Parsing response failed\n---\n%s\n---", e)
        return [{}]


def get_file_indices(
    picked_file: str, dataset_files: List[Dict[str, str]]
) -> List[int]:
    """
    Returns the index of the picked file in the dataset if the provided input is not an integer.

    Args:
        picked_file: glob pattern for files to be downloaded
        dataset_files:
    """
    try:
        idx = int(picked_file)
        return [idx]
    except ValueError:
        matched_indices = []
        for idx, dataset_file in enumerate(dataset_files):
            if fnmatch.fnmatch(dataset_file.get("title", ""), picked_file):
                matched_indices.append(idx)
    return matched_indices


def humanize_size(size: int) -> str:
    """
    Returns the file size (inp=bytes) in a human readable format

    Args:
        size: size in bytes

    Returns:
        size in human readable format
    """
    try:
        size = float(size)
    except:
        return size

    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(size) < 1000:
            return f"{size:.2f} {unit}"

        size /= 1000

    return f"{size:.2f} YB"


class Downloader:
    """
    A rudimentary URL downloader (like wget or curl) to demonstrate Rich progress bars.

    Adapted from [here](https://github.com/willmcgugan/rich/blob/master/examples/downloader.py)
    """

    def __init__(
        self, njobs: int = None, api_key: str = None, chunk_size: int = 1024 * 1024
    ):
        """
        Args:
            njobs: number of parallel downloads
            api_key: AIcrowd API Key
            chunk_size: no. of bytes to process in each iteration
        """
        self.progress_bar = TqdmProgressBar()

        if njobs is None:
            njobs = max(os.cpu_count() // 2, 1)

        self.njobs = njobs
        self.api_key = api_key
        # need this to indicate when the thread should die
        self.__active = False
        self.chunk_size = self.get_chunk_size(chunk_size)

    @staticmethod
    def get_chunk_size(chunk_size: int) -> int:
        """
        Set chunk size of 100 MB if running on Google colab

        Args:
            chunk_size: default size to return if not on colab
        """
        if is_google_colab_env():
            return 32 * 1024 * 1024
        else:
            return chunk_size

    def copy_url(self, filename: str, url: str, path: str) -> None:
        """
        Copy data from a url to a local file

        Args:
            task_id: rich progressbar task id
            url: url for the file to be downloaded
            path: path where the file will be saved
        """
        req = Request(url)
        if urlparse(url).netloc == DATASETS_HOST:
            req.add_unredirected_header("Authorization", f"Token {self.api_key}")
        response = urlopen(req)

        # This will break if the response doesn't contain content length
        task_id = self.progress_bar.add(
            filename=filename,
            total=int(response.info().get("Content-length")),
        )

        with open(path, "wb") as dest_file:
            for data in iter(partial(response.read, self.chunk_size), b""):
                # parent has quit, die
                if not self.__active:
                    return

                dest_file.write(data)
                self.progress_bar.update(progress_bar_id=task_id, step=len(data))
        self.progress_bar.close(progress_bar_id=task_id)

    def download(self, urls: Iterable[str], dest_dir: str):
        """
        Download multuple files to the given directory

        Args:
            urls: list of urls from which files are to be downloaded
            dest_dir: downloaded files will end up here
        """
        self.__active = True

        # https://stackoverflow.com/q/29177490
        #
        # The threads created by ThreadPoolExecutor are daemon
        # Normally, when running as CLI, the parent dies and so do the threads
        #
        # When this is run inside a notebook as a magic command, this doesn't happen
        #   which means that threads are
        #    - still downloading files
        #    - continuously updating the progress bars

        try:
            with ThreadPoolExecutor(max_workers=self.njobs) as pool:
                for url in urls:
                    filename = urlparse(url).path.split("/")[-1]
                    dest_path = os.path.join(dest_dir, filename)
                    pool.submit(self.copy_url, filename, url, dest_path)
        except KeyboardInterrupt:
            # the thread will read this
            self.__active = False
            raise



"""
Global constants file
"""

import os

RAILS_HOST = os.getenv("RAILS_HOST", "www.aicrowd.com")
RAILS_API_ENDPOINT = f"https://{RAILS_HOST}/api/v1"
AICROWD_API_ENDPOINT = os.getenv(
    "AICROWD_API_ENDPOINT", "https://aicrowd-api.aws-internal.k8s.aicrowd.com"
)

DATASETS_HOST = os.getenv("DATASETS_HOST", "datasets.aicrowd.com")


class LoginConstants:
    """
    Constants related to the login subcommand
    """

    CONFIG_KEY = "aicrowd_api_key"


class ChallengeConstants:
    """
    Constants related to challenge command
    """

    CONFIG_SECTION_NAME = "aicrowd-cli"
    CONFIG_ID_KEY = "challenge-id"
    CONFIG_SLUG_KEY = "challenge-slug"


class JupyterConfigConstants:
    ASSETS_DIR = "ASSETS_DIR"
    AICROWD_CONFIG = "AIcrowdConfig"



"""
Dataset subcommand
"""

import sys
from typing import List

import click

from aicrowd.contexts import (
    pass_config,
    ConfigContext,
    pass_challenge,
    ChallengeContext,
)
from aicrowd.utils import AliasedGroup


@click.group(name="dataset", cls=AliasedGroup)
def dataset_command():
    """
    View and download datasets
    """


@click.command(name="list")
@click.option("-c", "--challenge", type=str, help="Specify challenge explicitly")
@pass_challenge
@pass_config
def list_subcommand(
    config_ctx: ConfigContext, challenge_ctx: ChallengeContext, challenge: str
):
    """
    List the dataset files
    """
    from aicrowd.dataset import list_
    from aicrowd.dataset.exceptions import DatasetException

    try:
        list_(challenge, False, config_ctx, challenge_ctx)
    except DatasetException as e:
        click.echo(click.style(e.message, fg="red"))
        if e.fix:
            click.echo(click.style(e.fix, fg="yellow"))
        sys.exit(e.exit_code)


@click.command(name="download")
@click.option("-c", "--challenge", type=str, help="Specify challenge explicitly")
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(exists=True),
    help="Specify dataset download directory",
    default=".",
)
@click.option("-j", "--jobs", type=int, help="Number of files to download in parallel")
@click.argument("datasets", nargs=-1, type=str)
@pass_challenge
@pass_config
def download_subcommand(
    config_ctx: ConfigContext,
    challenge_ctx: ChallengeContext,
    challenge: str,
    output_dir: str,
    jobs: int,
    datasets: List[str],
):
    """
    Download dataset files
    """
    from aicrowd.dataset import download
    from aicrowd.dataset.exceptions import DatasetException

    try:
        download(challenge, output_dir, jobs, datasets, config_ctx, challenge_ctx)
    except DatasetException as e:
        click.echo(click.style(e.message, fg="red"))
        if e.fix:
            click.echo(click.style(e.fix, fg="yellow"))
        sys.exit(e.exit_code)


dataset_command.add_command(list_subcommand)
dataset_command.add_command(download_subcommand)



import os
import json

from aicrowd.notebook.helpers import delete_expressions_from_notebook

DEFAULT_CLEAN_EXPRESSIONS = [
    "API_KEY(\s+)?=(\s+)?[\"'].+[\"']",
    "^!aicrowd ",
]


def clean_notebook(file_path: str, output_path: str = None):
    """
    Remove sensitive data from the notebook and write the cleaned notebook
    in place

    Args:
        file_path: Path to the notebook
        output_path: Path to write the cleaned notebook to
    """
    print("Scrubbing API keys from the notebook...")
    clean_expressions = os.getenv("AICROWD_CLEAN_EXPRESSIONS")
    if clean_expressions is None:
        clean_expressions = DEFAULT_CLEAN_EXPRESSIONS
    else:
        clean_expressions = json.loads(clean_expressions)

    cleaned_nb = delete_expressions_from_notebook(clean_expressions, file_path)

    if output_path is None:
        output_path = file_path
    with open(output_path, "w") as fp:
        json.dump(cleaned_nb, fp)



"""
The dataset list subcommand
"""

import logging
from typing import List, Union

from rich import box
from rich.console import Console
from rich.table import Column, Table

from aicrowd.constants import ChallengeConstants
from aicrowd.contexts import ChallengeContext, ConfigContext
from aicrowd.dataset.exceptions import ChallengeNotFoundException
from aicrowd.dataset.helpers import (
    get_datasets,
    humanize_size,
    must_get_api_key,
    parse_cli_challenge,
)
from aicrowd.errors import INVALID_PARAMETER


def list_(
    challenge: str,
    return_list: bool = True,
    config_ctx: ConfigContext = ConfigContext(),
    challenge_ctx: ChallengeContext = ChallengeContext(),
) -> Union[None, List[dict]]:
    """
    Lists the datasets available for this challenge

    Args:
        challenge: one of

            - [`int`] challenge id
            - [`str`] challenge slug
            - [`str`] challenge url
        return_list: if true, will return the datasets instead of pretty printing
        config_ctx: CLI config
        challenge_ctx: Challenge config

    Returns:
        list of datasets if `return_list` was True; None otherwise
    """
    log = logging.getLogger()

    api_key = must_get_api_key(config_ctx)
    challenge_id = challenge_ctx.challenge.get(ChallengeConstants.CONFIG_ID_KEY)

    # couldn't get from config, try the --challenge option
    if challenge_id is None:
        challenge_id, _ = parse_cli_challenge(challenge, api_key)

    # still couldn't deduce challenge
    if challenge_id is None:
        log.error("Failed to parse challenge")
        raise ChallengeNotFoundException(
            "Challenge with the given details could not be found",
            exit_code=INVALID_PARAMETER,
        )

    datasets = get_datasets(challenge_id, api_key)
    log.info("Got %d datasets", len(datasets))

    if return_list:
        return datasets

    console = Console()
    table = Table(
        Column("#", max_width=2),
        "Title",
        "Description",
        Column(header="Size", justify="right"),
        title=f"Datasets for challenge #{challenge_id}",
        show_header=True,
        header_style="bold magenta",
        box=box.SQUARE,
    )

    empty_to_dash = lambda x: "-" if not x or len(x) == 0 else x

    key_mapper = [
        ("title", empty_to_dash),
        ("description", empty_to_dash),
        ("external_file_size", humanize_size),
    ]

    for i, ds in enumerate(datasets):
        vals = [str(i)]

        for key, mapper in key_mapper:
            vals.append(mapper(ds.get(key)))

        table.add_row(*vals)

    console.print(table, justify="left")



"""
The dataset download subcommand
"""

import logging
from typing import List

from aicrowd.dataset.list_cmd import list_
from aicrowd.dataset.helpers import Downloader, must_get_api_key, get_file_indices
from aicrowd.contexts import ConfigContext, ChallengeContext
from aicrowd.errors import INVALID_PARAMETER
from aicrowd.dataset.exceptions import DatasetNotFoundException


def download(
    challenge: str,
    output_dir: str,
    jobs: int,
    dataset_files: List[str],
    config_ctx: ConfigContext = ConfigContext(),
    challenge_ctx: ChallengeContext = ChallengeContext(),
):
    """
    Downloads the specified datasets available for this challenge

    Args:
        challenge: one of

            - [`int`] challenge id
            - [`str`] challenge slug
            - [`str`] challenge url
        output_dir: where to download files, cwd by default
        jobs: how many files to download in parallel, (no. of cores)/2 by default
        dataset_files: which ones to download?, all by default
        config_ctx: CLI config
        challenge_ctx: Challenge config

    Example:
        To download all datasets for "example-challenge" into the current dir, 4 in parallel

        ```python
        from aicrowd import dataset
        dataset.download("example-challenge", "./", 4, [])
        ```
    """
    log = logging.getLogger()
    api_key = must_get_api_key(config_ctx)

    datasets = list_(challenge, True, config_ctx, challenge_ctx)
    n_datasets = len(datasets)

    log.info("Got %d datasets", n_datasets)

    if len(dataset_files) == 0:
        dataset_files = list(range(len(datasets)))

    filtered_datasets = []

    dataset_file_indices = []
    for i in dataset_files:
        dataset_file_indices += get_file_indices(i, datasets)
    dataset_file_indices = list(set(dataset_file_indices))

    for file_index in dataset_file_indices:
        filtered_datasets.append(datasets[file_index])

    if len(filtered_datasets) == 0:
        raise DatasetNotFoundException(
            "No valid dataset found", "Run `aicrowd dataset list`", INVALID_PARAMETER
        )

    Downloader(jobs, api_key).download(
        map(lambda x: x.get("external_url"), filtered_datasets), output_dir
    )



"""
%aicrowd magic command
"""
import json
import os
import subprocess
from functools import wraps
from typing import Callable, List, Dict, Any

# this will never get called unless you are using ipython (or jupyter)
# since they will already have IPython installed, there's no need to add this as a dependency
# pylint: disable=E0401
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic

# pylint: enable=E0401
from aicrowd.cli import cli


class ExecutionError(Exception):
    """Exception thrown when a shell command exits with non zero code"""

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "ExecutionError: {0} ".format(self.message)
        else:
            return "ExecutionError: code execution failed"


def is_evaluation_env():
    """Returns true if running as a part of evaluation"""
    return os.getenv("AICROWD_IS_EVALUATING") is not None


def write_array_as_txt(array: List[str], file_path: str):
    """Write the elements of the input array to a file

    Args:
        array: List of elements to write to a file
        file_path: Path to the file that needs to be written
    """
    with open(file_path, "w") as fp:
        fp.write("\n".join(array))


def execute_command(command: str) -> int:
    """Capture the outputs from subprocess and send them to current process'
    stdout

    The output to subprocess' stdout will not be printed on the jupyter
    notebook. So we need to capture the output and print to current process'
    stdout.

    Args:
        command: Bash command to execute

    Returns:
        Exit code of the bash command
    """
    process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)

    def print_output(handle):
        if handle is not None:
            output = handle.read()
            if output:
                print(output.decode().strip())

    while True:
        if process.poll() is not None:
            break
        print_output(process.stdout)
        print_output(process.stderr)
    rc = process.poll()
    return rc


def override_evaluation_variables(user_ns: Dict[str, Any]):
    config = user_ns.get("AIcrowdConfig")
    if config is None:
        raise NameError("AIcrowdConfig is not defined")
    config.DATASET_PATH = os.getenv("DATASET_PATH")
    config.PREDICTIONS_PATH = os.path.join(
        os.getenv("OUTPUTS_DIR", "."),
        os.getenv("OUTPUT_FILE_NAME", "predictions.csv"),
    )
    for k, v in json.loads(os.getenv("AICROWD_CONFIG_ENV", "{}")):
        setattr(config, k, v)


def skip_during_evaluation(fn: Callable):
    """Executes notebook cell if not in evaluation mode

    Args:
        fn: Function to wrap
    """

    @wraps(fn)
    def _skipper(self, *args, **kwargs):
        if is_evaluation_env():
            override_evaluation_variables(self.shell.user_ns)
            return

        if len(args) == 2:
            cell = args[1]
        elif len(args) == 3:
            cell = args[2]
        else:
            cell = kwargs["cell"]

        self.shell.run_cell(cell)
        return fn(self, *args, **kwargs)

    return _skipper


def ignore_during_evaluation(fn: Callable):
    """Executes the method when not in evaluation mode

    Args:
        fn: Function to wrap
    """

    @wraps(fn)
    def _skipper(self, *args, **kwargs):
        if is_evaluation_env():
            override_evaluation_variables(self.shell.user_ns)
            return
        return fn(self, *args, **kwargs)

    return _skipper


@magics_class
class AIcrowdMagics(Magics):
    """AIcrowd magic commands"""

    @line_magic
    @ignore_during_evaluation
    def aicrowd(self, line: str):
        """Utility function to expose CLI commands as line magic commands

        Args:
            line: Text written next to the command
        """
        try:
            cli.main(line.split(" "), "%aicrowd")
        except Exception as e:
            print("An error occured:", e)
        except SystemExit:
            # Captures sys.exit from click
            pass

    @cell_magic
    @skip_during_evaluation
    def skip_aicrowd_evaluation(self, line: str, cell: str):
        """Skips the execution of a cell during evaluation

        Args:
            line: Text written next to the command
            cell: Contents of the cell excluding magic command
        """
        pass

    @cell_magic
    @skip_during_evaluation
    def pre_evaluation_steps(self, line: str, cell: str):
        """Skips the execution of a cell during evaluation

        Same as `skip_aicrowd_evaluation`. During the docker build stage
        we will get the contents of this cell and execute them. This is to
        cover complex use cases where it is not possible to get everything
        installed using single `apt install` or `pip install`.

        Args:
            line: Text next to the magic command
            cell: Contents of the cell excluding the magic command
        """
        pass

    @line_magic
    def save_assets(self, line: str):
        config = self.shell.user_ns.get("AIcrowdConfig")
        if config is None:
            raise NameError("`AIcrowdConfig` class is not defined")
        assets_dir = getattr(config, "ASSETS_DIR")
        if assets_dir is None:
            raise AttributeError("`ASSETS_DIR` should be defined on AIcrowdConfig")

    @line_magic
    def install_packages(self, line: str):
        """Install apt and pypi packages specified in `AIcrowdConfig` class

        Args:
            line: Text next to the magic command

        Example:
            To install wget using apt and numpy using pip, the user has to
            populate `AIcrowdConfig` as

            ```python
            class AIcrowdConfig:
                APT_PACKAGES = ["wget"]
                PIP_PACKAGES = ["numpy"]
            ```
        """
        config = self.shell.user_ns.get("AIcrowdConfig")
        if config is None:
            raise NameError("`AIcrowdConfig` class is not defined")

        if is_evaluation_env():
            config.DATASET_PATH = os.getenv("DATASET_PATH")
            config.PREDICTIONS_PATH = os.path.join(
                os.getenv("OUTPUTS_DIR", "."),
                os.getenv("OUTPUT_FILE_NAME", "predictions.csv"),
            )
            return

        apt_packages = getattr(config, "APT_PACKAGES", None)
        if apt_packages is not None:
            cmd = [
                "sudo apt -qq update",
                "sudo apt -qq install -y {}".format(" ".join(apt_packages)),
            ]
            if execute_command(" && ".join(cmd)) != 0:
                raise ExecutionError("Couldn't install apt packages")
            write_array_as_txt(apt_packages, "apt.txt")

        pip_packages = getattr(config, "PIP_PACKAGES", None)
        if pip_packages is not None:
            cmd = "pip install {}".format(" ".join(pip_packages))
            if execute_command(cmd) != 0:
                raise ExecutionError("Couldn't install pip packages")
            write_array_as_txt(pip_packages, "requirements.txt")


def load_ipython_extension(ipython):
    """Register our magic commands with ipython kernel

    The magic commands will be available as an extension amd can be loaded
    using

        %load_ext aicrowd.magic
    """
    ipython.register_magics(AIcrowdMagics)



"""
Subcommand to copy the assets to user's google drive
"""

import logging
import os
import shutil

import click

from aicrowd.submission.helpers import mount_google_drive
from aicrowd.utils.jupyter import is_google_colab_env


def load_assets(
    challenge: str,
    assets_dir: str,
):
    """
    Copy the assets from local FS to google drive. These assets will later be used
    in the inference notebooks.

    Args:
        challenge: Challenge identifier
        assets_dir: Path to the directory containing assets

    Raises:
        RuntimeError: If the command is triggered outside google colab notebook
    """
    log = logging.getLogger()
    if not is_google_colab_env():
        log.debug("Checking for google colab env")
        click.echo(
            click.style("This command should only be run on Google Colab Notebooks")
        )
        raise RuntimeError("Not running on google colab")

    log.debug("Mounting google drive")
    mount_path = mount_google_drive(
        mount_reason="Google drive is needed to store your assets"
    )
    log.debug("Copying assets to google drive")
    shutil.copytree(assets_dir, os.path.join(mount_path, "aicrowd", challenge))
    log.debug("Copied assets to google drive")



"""
Exceptions for dataset subcommand
"""

from aicrowd.exceptions import CLIException


class DatasetException(CLIException):
    """
    Base exception for the dataset subcommand
    """


class ChallengeNotFoundException(DatasetException):
    """
    The queried challenge was not found
    """


class DatasetNotFoundException(DatasetException):
    """
    The requested dataset could not be found
    """



"""
The submission create subcommand
"""

import logging

from aicrowd.contexts import ConfigContext, ChallengeContext
from aicrowd.constants import ChallengeConstants
from aicrowd.errors import INVALID_PARAMETER
from aicrowd.submission.helpers import (
    must_get_api_key,
    parse_cli_challenge,
    submit_file,
    bundle_notebook_submission,
)
from aicrowd.submission.exceptions import (
    ChallengeNotFoundException,
    InvalidChallengeDirException,
)


def create(
    challenge: str,
    file_path: str,
    description: str,
    print_links: bool = False,
    config_ctx: ConfigContext = ConfigContext(),
    challenge_ctx: ChallengeContext = ChallengeContext(),
    jupyter: bool = False,
):
    """
    Creates a submission on AIcrowd

    Considers both cases:

     - git submission
     - artifact (file based) submission

    if `-f/--file` is specified, they are submitting a file.
    So, default to artifact based submission

    Otherwise, default to gitlab based submission
      (should ignore challenge parameter in that case, pick up stuff from gitconfig)

    Args:
        challenge: one of

            - [`int`] challenge id
            - [`str`] challenge slug
            - [`str`] challenge url
        file_path: file to submit
        description: description for the submission
        print_links: print helpful links related to the submission
        config_ctx: CLI config
        challenge_ctx: Challenge config
        jupyter: Bundles jupyter notebook submission
    """
    log = logging.getLogger()

    api_key = must_get_api_key(config_ctx)
    challenge_id = challenge_ctx.challenge.get(ChallengeConstants.CONFIG_ID_KEY)

    # not in a valid challenge git directory and file not given
    if challenge_id is None and file_path is None:
        log.error("Not in a valid challenge git directory and file not given")
        raise InvalidChallengeDirException(
            "Please run this command from the challenge directory for git based submissions "
            + "or specify the file using -f/--file for artifact based submissions",
            exit_code=INVALID_PARAMETER,
        )

    # couldn't get from config, try the --challenge option
    if challenge_id is None:
        challenge_id, challenge_slug = parse_cli_challenge(challenge, api_key)

    # still couldn't deduce challenge
    if challenge_id is None:
        raise ChallengeNotFoundException(
            "Challenge with the given details could not be found",
            exit_code=INVALID_PARAMETER,
        )

    # Bundle the notebook with assets if colab flag is passed
    if jupyter:
        file_path = bundle_notebook_submission()

    if file_path is None:
        raise NotImplementedError("Git based submissions are not ready yet!")

    return submit_file(challenge_slug, file_path, description, api_key, print_links)



import os
import subprocess

from rich.console import Console

from aicrowd.constants import JupyterConfigConstants


def get_ipython_object():
    # this will never succeed unless you are using ipython (or jupyter)
    # since they will already have IPython installed, there's no need to
    #   add this as a dependency

    from IPython import get_ipython  # pylint: disable=import-error

    return get_ipython()


def is_jupyter():
    """
    Tries to guess whether it's running inside jupyter notebook or not
    """
    # first, check if get_ipython() itself is defined
    try:
        ip = get_ipython_object()
        # now check if it's running as kernel
        return ip.has_trait("kernel")
    except:
        return False


def is_google_colab_env() -> bool:
    """
    Checks if the CLI is running on Google colab notebook.
    """
    try:
        import google.colab  # pylint: disable=import-error

        return True
    except ImportError:
        return False


def mount_google_drive(
    mount_path: str = "/content/drive",
    mount_reason: str = "Your Google Drive will be mounted to access the colab notebook",
) -> str:
    """Mount user's Google Drive to a path

    Args:
        mount_path: Path in local FS where the user's GDrive will be mounted
        mount_reason: Message displayed on the terminal when attempting to mount

    Returns:
        Mount path for user's GDrive
    """
    if os.path.exists(mount_path):
        return mount_path

    from google.colab import drive  # pylint: disable=import-error

    console = Console()
    console.print("Mounting Google Drive :floppy_disk:", style="bold blue")
    console.print(mount_reason)
    drive.mount(mount_path)
    return mount_path


def get_aicrowd_config():
    config = get_ipython_object().user_ns.get(JupyterConfigConstants.AICROWD_CONFIG)
    if config is None:
        raise KeyError(
            "{} should be initiated before running the command".format(
                JupyterConfigConstants.AICROWD_CONFIG
            )
        )
    return config


def get_assets_dir():
    config = get_aicrowd_config()
    assets_dir = getattr(config, JupyterConfigConstants.ASSETS_DIR, None)
    if assets_dir is None:
        raise AttributeError(
            "{} should be defined in {}".format(
                JupyterConfigConstants.ASSETS_DIR, JupyterConfigConstants.AICROWD_CONFIG
            )
        )
    return assets_dir


def execute_command(command: str) -> int:
    """Capture the outputs from subprocess and send them to current process'
    stdout

    The output to subprocess' stdout will not be printed on the jupyter
    notebook. So we need to capture the output and print to current process'
    stdout.

    Args:
        command: Bash command to execute

    Returns:
        Exit code of the bash command
    """
    process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)

    def print_output(handle):
        if handle is not None:
            output = handle.read()
            if output:
                print(output.decode().strip())

    while True:
        if process.poll() is not None:
            break
        print_output(process.stdout)
        print_output(process.stderr)
    rc = process.poll()
    return rc



"""
Contains project version information
"""

import click


@click.command(name="version")
def version_command():
    """
    Prints AIcrowd CLI version
    """
    from aicrowd import __version__

    click.echo(click.style(str(__version__), bold=True))



"""
The challenge init subcommand
"""

import logging
from pathlib import Path

from git import Repo

from aicrowd.challenge.exceptions import ChallengeNotFoundException
from aicrowd.challenge.helpers import (
    create_challenge_dir,
    must_get_api_key,
    parse_cli_challenge,
    validate_challenge_dir,
)
from aicrowd.constants import ChallengeConstants
from aicrowd.contexts import ConfigContext
from aicrowd.errors import INVALID_PARAMETER


def init(
    challenge: str,
    base_dir: str,
    mkdir: bool,
    config_ctx: ConfigContext = ConfigContext(),
):
    """
    Setups the basic challenge files

    Args:
        challenge: one of

            - [`int`] challenge id
            - [`str`] challenge slug
            - [`str`] challenge url
        base_dir: directory for setting up challenge
        mkdir: should the base directory be created if it doesn't exist?
        config_ctx: CLI Config

    Example:
        To create a challenge directory named `example` for `example-challenge`

        ```python
        from aicrowd import challenge
        challenge.init("example-challenge", "example", True)
        ```
    """
    log = logging.getLogger()

    api_key = must_get_api_key(config_ctx)

    challenge_id, challenge_slug = parse_cli_challenge(challenge, api_key)

    if challenge_id is None or challenge_slug is None:
        log.error("Failed to parse challenge")
        raise ChallengeNotFoundException(
            "Challenge with the given details could not be found",
            exit_code=INVALID_PARAMETER,
        )

    # create the challenge directory
    if base_dir is None:
        challenge_dir = Path.cwd()
    else:
        challenge_dir = base_dir

    if mkdir:
        challenge_dir = create_challenge_dir(challenge_dir, challenge_slug)

    if not validate_challenge_dir(challenge_dir):
        raise ChallengeNotFoundException(
            "Challenge dir should be empty and not reside within an existing git repository.",
            INVALID_PARAMETER,
        )

    repo = Repo.init(challenge_dir)

    with repo.config_writer() as git_conf:
        git_conf.add_value(
            ChallengeConstants.CONFIG_SECTION_NAME,
            ChallengeConstants.CONFIG_ID_KEY,
            challenge_id,
        )
        git_conf.add_value(
            ChallengeConstants.CONFIG_SECTION_NAME,
            ChallengeConstants.CONFIG_SLUG_KEY,
            challenge_slug,
        )



"""
Challenge related tasks
"""

import logging
import os
import re
import subprocess
from pathlib import Path
from typing import Tuple, Union

import click
import git
import requests

from aicrowd.constants import AICROWD_API_ENDPOINT, RAILS_HOST, LoginConstants
from aicrowd.contexts import ConfigContext


def must_get_api_key(config_ctx: ConfigContext) -> str:
    """
    Checks whether API Key is defined in the config, prints a help message otherwise

    Args:
        config_ctx: CLI Config

    Returns:
        AIcrowd API Key
    """
    return config_ctx.config.get(
        LoginConstants.CONFIG_KEY,
        ensure_exists=True,
        help_msg=" ".join(
            [
                click.style("Please run", fg="red"),
                click.style("aicrowd login", fg="yellow"),
                click.style("to login.", fg="red"),
            ]
        ),
    )


def parse_cli_challenge(challenge: Union[int, str], api_key: str) -> Tuple[int, str]:
    """
    Parses the cli input to extract challenge

    Args:
        challenge: one of

            - [`int`] challenge id
            - [`str`] challenge slug
            - [`str`] challenge url
        api_key: aicrowd api key

    Returns:
        challenge_id, challenge_slug
    """
    log = logging.getLogger()

    if challenge is None:
        return None, None

    if challenge.isnumeric():
        # id was provided
        challenge_id = int(challenge)
        challenge_slug = get_slug_from_id(challenge_id, api_key)
    else:
        if challenge.startswith("http"):
            challenge_slug = get_slug_from_url(challenge)

            if challenge_slug is None:
                log.warning(
                    "Couldn't parse slug from given URL, assuming slug was provided"
                )
                # just in case the challenge slug is like "http-challenge"
                challenge_slug = challenge
        else:
            # assume that slug was provided
            challenge_slug = challenge

        challenge_id = get_id_from_slug(challenge_slug, api_key)

    return challenge_id, challenge_slug


def get_challenge_with_params(key: str, value: str, api_key: str) -> dict:
    """
    Queries AIcrowd API for challenge where key=value

    Args:
        key: database column for the challenge
        value: value of that key for the challenge
        api_key: aicrowd api key

    Returns:
        challenge information
    """
    log = logging.getLogger()

    r = requests.get(
        f"{AICROWD_API_ENDPOINT}/challenges/",
        params={key: value},
        headers={"Authorization": f"Token {api_key}"},
    )

    if not r.ok:
        log.error("Request to API failed\nReason: %s\nMessage: %s", r.reason, r.text)
        return {}

    try:
        return r.json()[0]
    except Exception as e:
        log.error("Parsing response failed\n%s", e)
        return {}


def get_id_from_slug(slug: str, api_key: str) -> int:
    """
    Gets the challenge id from slug (if valid)

    Args:
        slug: challenge slug
        api_key: aicrowd api key

    Returns:
        challenge id
    """
    return get_challenge_with_params("slug", slug, api_key).get("id")


def get_slug_from_id(challenge_id: int, api_key: str) -> str:
    """
    Gets the challenge slug from id (if valid)

    Args:
        challenge_id: challenge id
        api_key: aicrowd api key

    Returns:
        challenge slug
    """
    return get_challenge_with_params("id", challenge_id, api_key).get("slug")


def get_slug_from_url(url: str) -> str:
    """
    Extracts slug from challenge url

    Args:
        url: challenge url

    Returns:
        challenge slug
    """
    log = logging.getLogger()

    slug_re = re.compile(f"https://{RAILS_HOST}/challenges/([^/]*)")
    slug_match = slug_re.match(url)

    if slug_match is None:
        log.error("Wrong url given? No match found")
        return None

    return slug_match.groups()[0]


def create_challenge_dir(challenge_base_dir: Path, challenge_slug: str) -> str:
    """
    Creates a new challenge directory with the help of slug

    Args:
        challenge_base_dir: base directory for creating the challenge
        challenge_slug:

    Returns:
        path to the created directory
    """
    log = logging.getLogger()
    dir_path = os.path.join(challenge_base_dir, challenge_slug)

    if os.path.exists(dir_path):
        suffix = 0
        log.warning(
            "Challenge directory %s already exists. A new one will be created", dir_path
        )

        while os.path.exists(f"{dir_path}-{suffix}"):
            suffix += 1

        dir_path += f"-{suffix}"

    Path(dir_path).mkdir(parents=True)
    return dir_path


def validate_challenge_dir(challenge_dir: str) -> bool:
    """
    Checks whether the challenge_dir is an empty directory outside git repo

    Args:
        challenge_dir: directory to be used for setting up challenge

    Returns:
        is challenge_dir valid?
    """
    log = logging.getLogger()

    if os.listdir(challenge_dir):
        log.error("Challenge dir is not empty")
        return False

    try:
        git.Repo(search_parent_directories=True)
        # if its a git repo, it's not valid
        return False
    except git.InvalidGitRepositoryError:
        return True



"""
Top level helpers
"""
import os
import subprocess
import uuid
from typing import Any

import click
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from tqdm.auto import tqdm

from aicrowd.utils.jupyter import is_jupyter


def is_subsequence(s1: str, s2: str) -> bool:
    """
    returns whether s1 is a subsequence of s2 or not

    Args:
        s1: probable subsequence of s2
        s2: string

    Returns:
        is s1 a subsequence of s2
    """
    i, j = 0, 0
    n1, n2 = len(s1), len(s2)

    while i < n1 and j < n2:
        # if matched, advance in both
        if s1[i] == s2[j]:
            i += 1

        # otherwise, try next char in s2
        j += 1

    # was s1 fully matched?
    return i == n1


class AliasedGroup(click.Group):
    """
    Click group allowing using prefix of command instead of the whole command
    """

    def get_command(self, ctx, cmd_name):
        """
        Returns a command of which the given word is a subsequence
        """
        rv = click.Group.get_command(self, ctx, cmd_name)

        if rv is not None:
            return rv

        matches = list(
            filter(
                lambda cmd_orig: is_subsequence(cmd_name, cmd_orig),
                self.list_commands(ctx),
            )
        )

        if not matches:
            return None

        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])

        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")
        return None


def zip_fs_path(fs_path: str, target_zip_path: str):
    fs_path = fs_path.replace(".zip", "")
    if not os.path.exists(fs_path):
        raise FileNotFoundError(f"The path {fs_path} does not exist")
    if (
        subprocess.run(f"zip -r '{target_zip_path}' '{fs_path}'", shell=True).returncode
        != 0
    ):
        raise Exception(f"Failed to zip {fs_path}")


class ProgressBar:
    def add(self, filename: str, total: int):
        raise NotImplementedError

    def update(self, progress_bar_id: Any, step: int):
        raise NotImplementedError

    def close(self, progress_bat_id: Any):
        raise NotImplementedError


class RichProgressBar(ProgressBar):
    def __init__(self):
        self.progress = Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
        )
        self.progress.console.is_jupyter = is_jupyter()
        self._tasks = []

    def add(self, filename: str, total: int, **kwargs):
        task_id = self.progress.add_task(
            "download",
            filename=filename,
            total=total,
            start=True,
            refresh=True,
            **kwargs,
        )
        self._tasks.append(task_id)
        return task_id

    def update(self, progress_bar_id: TaskID, step: int):
        self.progress.update(progress_bar_id, advance=step, refresh=True)

    def close(self, progress_bar_id: TaskID):
        pass


class TqdmProgressBar(ProgressBar):
    def __init__(self):
        self.bars = {}

    def add(self, filename: str, total: int, **kwargs):
        bar_id = uuid.uuid4()
        self.bars[bar_id] = tqdm(
            desc=filename, total=total, unit="B", unit_scale=True, **kwargs
        )
        return bar_id

    def update(self, progress_bar_id: Any, step: int, **kwargs):
        self.bars[progress_bar_id].update(step, **kwargs)

    def close(self, progress_bar_id: str):
        self.bars[progress_bar_id].close()



#!/usr/bin/env python

"""
Entrypoint for the CLI application
"""

import logging

import click
from rich.logging import RichHandler
from rich import traceback

from aicrowd.contexts import pass_config, ConfigContext
from aicrowd.utils import AliasedGroup

from aicrowd.cmd import (
    challenge_command,
    dataset_command,
    login_command,
    submission_command,
    version_command,
    notebook_command,
)

traceback.install()

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG,
}  #: a mapping of `verbose` option counts to logging levels


# pylint: disable=no-value-for-parameter
@click.group(cls=AliasedGroup)
@click.option("-v", "--verbose", count=True, help="Enable verbose output.")
@click.option("--config-path", help="Path to config file", type=click.Path())
@pass_config
def cli(
    config_context: ConfigContext,
    verbose: int,
    config_path: str,
):
    """
    AIcrowd CLI
    """
    if verbose > 0:
        level = LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else logging.DEBUG
    else:
        level = logging.CRITICAL

    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(show_path=False, show_time=False)],
    )

    # load config from path
    config_context.config.load(config_path)


cli.add_command(challenge_command)
cli.add_command(dataset_command)
cli.add_command(login_command)
cli.add_command(submission_command)
cli.add_command(version_command)
cli.add_command(notebook_command)


if __name__ == "__main__":
    cli()



"""
Subcommand to copy the assets to user's google drive
"""

import logging
import os

import click

from aicrowd.submission.helpers import delete_and_copy_dir
from aicrowd.utils.jupyter import (
    is_google_colab_env,
    mount_google_drive,
    get_assets_dir,
)


def copy_assets(
    challenge: str,
    colab_to_drive: bool = True,
):
    """
    Copy the assets from local FS to google drive and vice-versa. These assets will later be used
    in the inference notebooks.

    Args:
        challenge: Challenge identifier
        colab_to_drive: Direction of the copy
    """
    log = logging.getLogger()
    if not is_google_colab_env():
        log.debug("Checking for google colab env")
        click.echo(
            click.style("This command should only be run on Google Colab Notebooks")
        )
        raise RuntimeError("Not running on google colab")

    assets_dir = get_assets_dir()
    if colab_to_drive and not os.path.exists(assets_dir):
        raise FileNotFoundError(f"ASSETS_DIR: {assets_dir} does not exist")

    log.debug("Mounting google drive")
    mount_path = mount_google_drive(
        mount_reason="Google drive is needed to store your assets"
    )
    log.debug("Copying assets to google drive")
    drive_path = os.path.join(mount_path, "My Drive/aicrowd/challenges", challenge)
    if colab_to_drive:
        delete_and_copy_dir(assets_dir, drive_path)
    else:
        delete_and_copy_dir(drive_path, assets_dir)
    log.debug("Copied assets to google drive")



"""
Contains all click contexts
"""

from aicrowd.contexts.challenge import Challenge
from aicrowd.contexts.config import CLIConfig

from aicrowd.contexts.context import (
    ConfigContext,
    pass_config,
    ChallengeContext,
    pass_challenge,
)



"""
Login command
"""

import logging

import click

from aicrowd.constants import LoginConstants
from aicrowd.contexts import ConfigContext
from aicrowd.errors import INVALID_CREDENTIALS

from aicrowd.auth.helpers import verify_api_key
from aicrowd.auth.exceptions import CredentialException


def login(api_key: str, config_context: ConfigContext = ConfigContext()):
    """
    Log in using AIcrowd API Key

    Args:
        api_key: AIcrowd API Key
        config_context: CLI Config
    """
    config = config_context.config
    log = logging.getLogger()

    log.info("Verifying API Key...")

    if not verify_api_key(api_key):
        log.error("Invalid API Key provided")
        raise CredentialException(
            "Invalid API Key provided", exit_code=INVALID_CREDENTIALS
        )

    log.info("API Key verified")
    click.echo(click.style("API Key valid", fg="green"))

    current_api_key = config.get(LoginConstants.CONFIG_KEY)

    # if not defined or an older value, update
    if current_api_key != api_key:
        log.info("API Key will be (over) written")
        config.set(LoginConstants.CONFIG_KEY, api_key)

    config.write()
    click.echo(click.style("Saved API Key successfully!", fg="green"))



"""
Challenge subcommand
"""

import sys

import click

from aicrowd.contexts import pass_config, ConfigContext
from aicrowd.utils import AliasedGroup


@click.group(name="challenge", cls=AliasedGroup)
def challenge_command():
    """
    Setup a challenge
    """


@click.command(name="init")
@click.argument("challenge", type=str)
@click.option(
    "-d",
    "--base-dir",
    type=click.Path(exists=True),
    help="Base directory for storing the challenge",
)
@click.option(
    "--mkdir",
    is_flag=True,
    help="Create a new directory for challenge inside current directory",
)
@pass_config
def init_subcommand(
    config_ctx: ConfigContext, challenge: str, base_dir: str, mkdir: bool
):
    """
    Setups basic challenge files
    """
    from aicrowd.challenge import init
    from aicrowd.challenge.exceptions import ChallengeException

    try:
        init(challenge, base_dir, mkdir, config_ctx)
    except ChallengeException as e:
        click.echo(click.style(e.message, fg="red"))
        if e.fix:
            click.echo(click.style(e.fix, fg="yellow"))
        sys.exit(e.exit_code)


challenge_command.add_command(init_subcommand)



import os
import shutil

import click

from aicrowd.utils import AliasedGroup


@click.group(name="notebook", cls=AliasedGroup)
def notebook_command():
    """
    Notebook specific operations
    """
    pass


@click.command(name="clean")
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    required=True,
    help="Path to the notebook",
)
def clean_subcommand(file: str):
    """
    Remove sensitive information from the notebook

    The default behaviour is to match `AICROWD_API_KEY(\s+)?=(\s+)?".+"` and
    remove the lines containing the regex. To match custom expressions, set a
    list of regex in `AICROWD_CLEAN_EXPRESSIONS` environment variable.
    """
    from aicrowd.notebook.clean import clean_notebook

    clean_notebook(file_path=file)


@click.command(name="split")
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    required=True,
    help="Path to the notebook",
)
@click.option(
    "-o",
    "--output-dir",
    default="submission",
    help="Directory to place the split notebooks",
)
def split_subcommand(file: str, output_dir: str):
    """
    Split the notebook into training and prediction notebooks
    """
    from aicrowd.notebook.split import split_notebook

    split_notebook(file, output_dir)


@click.command(name="submit")
@click.option(
    "-c", "--challenge", required=True, help="Challenge to submit the file for"
)
@click.option("--description", default="", help="Description for the submission")
@click.option("-a", "--assets-dir", required=True, help="Path to assets dir")
@click.option("-o", "--output", default="submission.zip", help="Output file")
@click.option("-n", "--notebook-name", default=None, help="Name of the notebook to use")
@click.option("--no-verify", is_flag=True, help="Skip submission verification")
@click.option("--dry-run", is_flag=True, help="Verify submission but don't submit")
def submit_subcommand(
    challenge: str,
    description: str,
    assets_dir: str,
    output: str = "submission.zip",
    notebook_name: str = None,
    no_verify: bool = False,
    dry_run: bool = False,
):
    """
    Submit the notebook to AIcrowd
    """
    from aicrowd.notebook.submit import create_submission

    create_submission(
        challenge,
        description,
        assets_dir,
        output,
        notebook_name,
        no_verify,
        dry_run,
    )


@click.command(name="extract")
@click.option("-f", "--file", type=click.Path(exists=True), help="Path to the notebook")
@click.option("--start", help="Extract/Remove cells from this block")
@click.option("--end", help="Extract/Remove cells till this block")
@click.option(
    "--remove",
    is_flag=True,
    help="Removes the content between start and end instead of extracting them",
)
def extract_subcommand(file: str, start: str, end: str, remove: bool = False):
    """
    Extract/Remove cells from notebook between start and end
    """
    from aicrowd.notebook.split import extract_partial_notebook
    from aicrowd.notebook.helpers import read_notebook, write_notebook

    nb = read_notebook(file)
    partial_nb = extract_partial_notebook(
        nb=nb, start_block=start, end_block=end, remove=remove
    )
    write_notebook(file_path=os.path.join(os.getcwd(), file), nb=partial_nb)


@click.command(name="copy")
@click.option("-o", "--output", required=True, help="Path to copy the notebook to")
@click.option(
    "--clean", is_flag=True, help="Remove sensitive information from notebook"
)
def copy_subcommand(output: str, clean: bool = False):
    """
    Copies the active notebook to a path
    """
    from aicrowd.notebook.helpers import get_notebook_path
    from aicrowd.notebook.clean import clean_notebook

    shutil.copy2(get_notebook_path(), output)
    if clean:
        clean_notebook(output)


notebook_command.add_command(clean_subcommand)
notebook_command.add_command(split_subcommand)
notebook_command.add_command(submit_subcommand)
notebook_command.add_command(extract_subcommand)
notebook_command.add_command(copy_subcommand)



"""
Exceptions that the CLI raises

All other exceptions will inherit from `CLIException` so just catching that should be enough
"""

from aicrowd import errors


class CLIException(Exception):
    """
    Base class for all exceptions raised by CLI
    """

    def __init__(
        self, message: str, fix: str = None, exit_code: int = errors.UNKNOWN_ERROR
    ):
        """
        Args:
            message: this will be printed to the console
            fix: (if defined) what steps can be taken to fix the error
            exit_code: what the exit code should be if this exception was raised
        """
        super().__init__(message)

        self.message = message
        self.fix = fix
        self.exit_code = exit_code

    def __repr__(self):
        return f"{type(self).__name__}\nMessage: {self.message}\nFix?: {self.fix}"



"""
Exceptions for login
"""

from aicrowd.exceptions import CLIException


class LoginException(CLIException):
    """
    Base exception for login command
    """


class CredentialException(LoginException):
    """
    credential verification exception
    """



"""
Lists all error names and their exit codes
"""

EXIT_SUCCESS = 0

# exit codes 1 and 2 are used internally by click

INVALID_FILE = 3
INVALID_CREDENTIALS = 4
INVALID_PARAMETER = 5

CONFIG_NOT_SET = 6

UNKNOWN_ERROR = 7



"""
For click contexts
"""

import click

from aicrowd.contexts import CLIConfig, Challenge


class ConfigContext:
    """
    Contains:
     - CLIConfig
    """

    def __init__(self):
        self.config: CLIConfig = CLIConfig()

    def __repr__(self):
        return "ConfigContext()"


class ChallengeContext:
    """
    Contains information about the challenge for cwd
    """

    def __init__(self):
        self.challenge: Challenge = Challenge()

    def __repr__(self):
        return "ChallengeContext()"


pass_config = click.make_pass_decorator(ConfigContext, ensure=True)
pass_challenge = click.make_pass_decorator(ChallengeContext, ensure=True)



"""
Exceptions for challenge subcommand
"""

from aicrowd.exceptions import CLIException


class ChallengeException(CLIException):
    """
    Base exception for the challenge subcommand
    """


class ChallengeNotFoundException(ChallengeException):
    """
    The queried challenge was not found
    """



"""
Submission related tasks
"""
import json
import logging
import os
import re
import shutil
from pathlib import Path
from typing import Tuple, List, Dict, Any

import requests
from urllib.parse import unquote
from requests_toolbelt import MultipartEncoderMonitor
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.table import Table

# pylint: disable=unused-import
from aicrowd.challenge.helpers import must_get_api_key, parse_cli_challenge

# pylint: enable=unused-import
from aicrowd.constants import RAILS_API_ENDPOINT, RAILS_HOST
from aicrowd.utils.jupyter import (
    is_jupyter,
    is_google_colab_env,
    mount_google_drive,
    get_aicrowd_config,
)
from aicrowd.utils.utils import zip_fs_path
from aicrowd.submission.exceptions import (
    SubmissionFileException,
    SubmissionUploadException,
)


def calculate_min_table_width(table: Table):
    """
    Calculates minumum width needed to display table text properly

    Args:
        table: rich table
    """
    width = sum(table._calculate_column_widths(Console(), 200)) + table._extra_width
    term_width = shutil.get_terminal_size().columns
    return max(width, term_width)


def get_upload_details(challenge_slug: str, api_key: str) -> Tuple[dict, bool]:
    """
    Contacts AIcrowd website for getting presigned url for uploading

    Args:
        challenge_slug: challenge slug
        api_key: AIcrowd API Key

    Returns:
        the data and whether the request was successful or not
    """
    log = logging.getLogger()

    r = requests.get(
        f"{RAILS_API_ENDPOINT}/submissions",
        params={"challenge_id": challenge_slug},
        headers={"Authorization": f"Token token={api_key}"},
        allow_redirects=False,
    )

    # temporary hack until /api stops redirecting
    if r.status_code // 100 == 3:
        redirected_to = r.headers["Location"]
        expected_redirect = f"https://{RAILS_HOST}"

        if redirected_to.startswith(expected_redirect):
            redirected_to = redirected_to[len(expected_redirect) :]
        else:
            # got redirected to a weird place
            logging.error(
                "Unexpected redirect location: %s\nExpected %s",
                redirected_to,
                expected_redirect,
            )
            return {}, False

        challenge_problem = re.match(
            r"/challenges/([^/]*)/problems/([^/]*)/.*", redirected_to
        ).groups()
        logging.info("[metachallenge?] Got redirected to %s", redirected_to)

        # inform caller about meta_challenge
        return {
            "meta_challenge": True,
            "meta_challenge_id": challenge_problem[0],
            "challenge_id": challenge_problem[1],
        }, True

    try:
        resp = r.json()

        if not r.ok or not resp.get("success"):
            log.error(
                "Request to API failed\nReason: %s\nMessage: %s", r.reason, r.text
            )
            return resp, False

        return resp.get("data"), True
    except Exception as e:
        log.error("Error while extracting details: %s", e)
        return {}, False


class S3Uploader:
    """
    Upload files to s3 with progress bar
    """

    def __init__(self, host: str, fields: dict, file_path: str):
        """
        Args:
            host: s3 host to upload to
            fields: s3 related fields
            file_path: the file to upload
        """
        self.host = host
        self.fields = fields

        file_path = Path(file_path)

        self.file_name = file_path.name
        self.file_size = file_path.stat().st_size

        self.fields["key"] = self.fields["key"].replace("${filename}", self.file_name)
        self.fields["file"] = (self.file_name, file_path.open("rb"))

        self.progress = Progress(
            TextColumn("[bold blue]{task.fields[file_name]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
        )
        self.progress.console.is_jupyter = is_jupyter()
        self.task_id = self.progress.add_task(
            "upload", file_name=self.file_name, start=False, total=self.file_size
        )

    def track_progress(self, monitor: MultipartEncoderMonitor):
        """
        shows progress bar showing how much has been uploaded

        Args:
            monitor: requests_toolbelt Monitor
        """
        self.progress.update(self.task_id, completed=monitor.bytes_read, refresh=True)

    def upload(self):
        """
        upload file to s3
        """
        self.progress.start_task(self.task_id)
        m = MultipartEncoderMonitor.from_fields(
            self.fields, callback=self.track_progress
        )

        with self.progress:
            return (
                requests.post(
                    self.host, data=m, headers={"Content-Type": m.content_type}
                ),
                self.fields["key"],
            )


def upload_submission(host: str, fields: dict, file_path: str) -> str:
    """
    uploads a file to s3 using presigned url details

    Args:
        host: s3 host to upload to
        fields: s3 related keys
        file_path: the file to be uploaded

    Returns:
        whether it was successful or not
    """
    log = logging.getLogger()

    r, s3_key = S3Uploader(host, fields, file_path).upload()

    if not r.ok:
        log.error(
            "Couldn't upload file to s3\nReason: %s\nMessage: %s", r.reason, r.text
        )
        return None

    return s3_key


def notify_rails_upload(
    challenge_slug: str,
    submitted_url: str,
    api_key: str,
    description: str,
    problem_slug: str = None,
) -> dict:
    """
    notify rails about the uploaded file on s3

    Args:
        challenge_slug: challenge slug
        submitted_url: the url to which the submitted file was uploaded to
        description: submission description
        problem_slug: Used when submitting to a meta challenge

    Returns:
        submission details from AIcrowd API
    """
    log = logging.getLogger()

    if problem_slug is None:
        payload = {"challenge_id": challenge_slug}
    else:
        payload = {"meta_challenge_id": challenge_slug, "challenge_id": problem_slug}

    payload.update(
        {
            "description": description,
            "submission_files": [{"submission_file_s3_key": submitted_url}],
        }
    )

    r = requests.post(
        f"{RAILS_API_ENDPOINT}/submissions",
        json=payload,
        headers={"Authorization": f"Token token={api_key}"},
    )

    if not r.ok:
        log.error("Request to API failed\nReason: %s\nMessage: %s", r.reason, r.text)
        return {}

    try:
        return r.json()
    except Exception as e:
        log.error("Couldn't json-decode rails response")
        return {}


def print_submission_links(challenge_slug: str, problem_slug: str, submission_id: int):
    """
    prints helpful links related to the submission

    Args:
        challenge_slug: challenge slug
        problem_slug: when submitting to a meta challenge
        submission_id: rails submission id
    """
    if submission_id is None:
        return

    if problem_slug:
        challenge_url = (
            f"https://{RAILS_HOST}/challenges/{challenge_slug}/problems/{problem_slug}"
        )
    else:
        challenge_url = f"https://{RAILS_HOST}/challenges/{challenge_slug}"

    submission_base_url = f"{challenge_url}/submissions"

    table = Table(title="Important links", show_header=False, leading=1, box=box.SQUARE)
    table.add_column(justify="right")
    table.add_column(overflow="fold")

    table.add_row("This submission", f"{submission_base_url}/{submission_id}")
    table.add_row("All submissions", f"{submission_base_url}?my_submissions=true")
    table.add_row("Leaderboard", f"{challenge_url}/leaderboards")
    table.add_row(
        "Discussion forum", f"https://discourse.aicrowd.com/c/{challenge_slug}"
    )
    table.add_row("Challenge page", f"{challenge_url}")

    width = calculate_min_table_width(table)

    console = Console(width=width)
    table.min_width = width
    console.print(Panel("[bold]Successfully submitted!"), justify="center")
    console.print(table)


def submit_file(
    challenge_slug: str,
    file_path: str,
    description: str,
    api_key: str,
    print_links: bool,
) -> dict:
    """
    Submits a file given it's path and challenge_slug with given description

    Args:
        challenge_slug: challenge slug
        file_path: path to the file to be uploaded
        description: description for the submission
        api_key: AIcrowd API Key
        print_links: should helpful links be printed?

    Returns:
        a message from AIcrowd API
    """
    log = logging.getLogger()
    problem_slug = None

    if not Path(file_path).is_file():
        raise SubmissionFileException(f"Bad file {file_path}")

    s3_presigned_details, success = get_upload_details(challenge_slug, api_key)

    if s3_presigned_details.get("meta_challenge"):
        challenge_slug = s3_presigned_details.get("meta_challenge_id")
        problem_slug = s3_presigned_details.get("challenge_id")

        s3_presigned_details, success = get_upload_details(challenge_slug, api_key)

    if not success:
        log.error(
            "Error in getting presigned url for s3 upload: %s",
            s3_presigned_details.get("message"),
        )
        raise SubmissionUploadException(s3_presigned_details.get("message"))

    s3_key = upload_submission(
        s3_presigned_details["url"], s3_presigned_details["fields"], file_path
    )
    if s3_key is None:
        raise SubmissionUploadException(
            "Couldn't submit file. Please recheck the files and details provided"
        )

    response = notify_rails_upload(
        challenge_slug, s3_key, api_key, description, problem_slug
    )
    if not response.get("success"):
        raise SubmissionUploadException(
            response.get("message", "Error in notifying rails about the uploaded file")
        )

    if print_links:
        print_submission_links(
            challenge_slug, problem_slug, response.get("data", {}).get("submission_id")
        )

    return response.get("data")


def bundle_notebook_submission() -> str:
    """
    Bundle submission files along with the colab/jupyter notebook

    Any additional files in `AIcrowdConfig.ASSETS_DIR` will be added to the
    submission zip file.

    Returns:
        Path to the submission zip file
    """
    console = Console()

    if not is_jupyter():
        raise EnvironmentError("Not running in jupyter notebook")

    config = get_aicrowd_config()
    assets_dir = getattr(config, "ASSETS_DIR", None)
    if assets_dir is None:
        console.print("WARNING: No assets directory provided", style="bold red")

    target_zip_path = "submission"
    zip_assets(assets_dir, target_zip_path)

    if is_google_colab_env():
        zip_colab_notebook(target_zip_path)
    else:
        zip_jupyter_notebook(target_zip_path)

    return target_zip_path + ".zip"


def zip_assets(assets_dir: str, target_zip_path: str):
    """
    Zip the files under assets directory

    Args:
        assets_dir: Directory containing submission assets
        target_zip_path: Path to zip file to add the entries to
    """
    console = Console()
    if not os.path.exists(assets_dir):
        os.mkdir(assets_dir)
        console.print("WARNING: Assets dir is empty", style="bold red")
    zip_fs_path(assets_dir, target_zip_path)


def zip_jupyter_notebook(
    target_zip_path: str, submission_notebook_path: str = "submission.ipynb"
):
    """
    Add the jupyter notebook to the submission zip file

    The notebook will be available on the local file system which won't be
    the case colab notebook. So this should be used with colab notebook.

    Args:
        target_zip_path: Path to zip file to add the notebook to
        submission_notebook_path: Path to the submission notebook
    """
    notebook_path = get_jupyter_notebook_path()
    cleaned_nb = pop_source_line(
        os.path.join(notebook_path),
        ['API_KEY(\s+)?=(\s+)?".+"'],
    )
    with open(submission_notebook_path, "w") as fp:
        json.dump(cleaned_nb, fp)
    zip_fs_path(
        submission_notebook_path,
        target_zip_path,
    )


def get_jupyter_notebook_path():
    """
    Get the path to jupyter notebook from the sessions endpoint

    This assumes that jupyter is running without authentication on port 8888

    Returns:
        Path to the jupyter notebook
    """
    # TODO: Dynamically detect the endpoint
    # TODO: Support with notebooks with auth
    proxies = {
        "http": None,
        "https": None,
    }
    session_info = requests.get(
        "http://localhost:8888/api/sessions", proxies=proxies
    ).json()
    check_session_count(session_info)
    return unquote(session_info[0]["path"])


def zip_colab_notebook(target_zip_path: str):
    """
    Add the colab notebook to the submission zip file

    We do not have direct access to the colab notebook on the local FS. So,
    we mount the user's Google Drive and copy the notebook from there.

    Args:
        target_zip_path: Path to zip file to add the notebook to
    """
    mount_path = mount_google_drive()
    submission_notebook_path = "submission.ipynb"
    cleaned_nb = pop_source_line(
        os.path.join(
            mount_path, "My Drive", "Colab Notebooks", get_colab_notebook_name()
        ),
        ['API_KEY(\s+)?=(\s+)?".+"'],
    )
    with open(submission_notebook_path, "w") as fp:
        json.dump(cleaned_nb, fp)
    zip_fs_path(
        submission_notebook_path,
        target_zip_path,
    )


def pop_source_line(file_path: str, expressions: List[str]) -> Dict[str, Any]:
    """Remove certain code lines matching given regular expressions from the
    notebook

    Args:
        file_path: Submission notebook path
        expressions: List of regular expressions to match

    Returns:
        New notebook with matched lines removed
    """
    with open(file_path) as fp:
        nb = json.load(fp)

    for _cell in nb["cells"]:
        # Match the lines only in code blocks
        if _cell["cell_type"] == "code":
            source_code = []
            for _code_line in _cell["source"]:
                matched = False
                for _expr in expressions:
                    if re.search(_expr, _code_line):
                        matched = True
                        break
                if not matched:
                    source_code.append(_code_line)
            _cell["source"] = source_code
    return nb


def get_colab_notebook_name() -> str:
    """Get name of the current notebook

    Returns:
        Name of the current colab notebook
    """
    sesssion_info = requests.get("http://172.28.0.2:9000/api/sessions").json()
    check_session_count(sesssion_info)
    return unquote(sesssion_info[0]["name"])


def check_session_count(sesssion_info):
    if len(sesssion_info) == 0:
        raise EnvironmentError("Please start the notebook with a kernel")
    if len(sesssion_info) > 1:
        raise EnvironmentError(
            "Found more than one kernel running. Please close other notebooks and kernels and try again."
        )


def delete_and_copy_dir(src: str, dst: str):
    """
    Delete if the src exists and copy files from src to dst

    Args:
        src: Source path
        dst: Destination path
    """
    if os.path.exists(dst):
        shutil.rmtree(dst)
    shutil.copytree(src, dst)



#!/usr/bin/env python
"""
Build and install the CLI
"""

from pathlib import Path

from setuptools import setup, find_packages

import versioneer

README = Path("README.md").read_text()
REPO_URL = "https://gitlab.aicrowd.com/aicrowd/aicrowd-cli"

current_dir = Path(__file__).resolve().parent
version = versioneer.get_version()

with open(str(current_dir / "requirements" / "build.txt")) as reqs_file:
    reqs = reqs_file.read().split()

setup(
    name="aicrowd-cli",
    description="A CLI app to interact with AIcrowd",
    long_description=README,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    version=version,
    cmdclass=versioneer.get_cmdclass(),
    install_requires=reqs,
    entry_points={
        "console_scripts": [
            "aicrowd=aicrowd.cli:cli",
        ]
    },
    python_requires=">=3.6",
    license="MIT",
    author="Yoogottam Khandelwal",
    author_email="yoogottamk@outlook.com",
    url=REPO_URL,
    download_url=f"{REPO_URL}/-/archive/{version}/aicrowd-cli-{version}.tar.gz",
    keywords=["AIcrowd", "CLI"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
)



"""
Exceptions for submission subcommand
"""

from aicrowd.exceptions import CLIException


class SubmissionException(CLIException):
    """
    Base exception for the challenge subcommand
    """


class ChallengeNotFoundException(SubmissionException):
    """
    The queried challenge was not found
    """


class InvalidChallengeDirException(SubmissionException):
    """
    The command was given an invalid challenge directory
    """


class SubmissionFileException(SubmissionException):
    """
    The file given for submission was invalid
    """


class SubmissionUploadException(SubmissionException):
    """
    Something went wrong while uploading the file
    """



"""
Contains all commands
"""

from aicrowd.cmd.challenge import challenge_command
from aicrowd.cmd.dataset import dataset_command
from aicrowd.cmd.login import login_command
from aicrowd.cmd.version import version_command
from aicrowd.cmd.submission import submission_command
from aicrowd.cmd.notebook import notebook_command



"""
Auth related tasks
"""

import logging
import requests

from aicrowd.constants import RAILS_API_ENDPOINT


def verify_api_key(api_key: str) -> bool:
    """
    Verifies if the API Key is valid or not

    Args:
        api_key: AIcrowd API Key

    Returns:
        True if API Key valid, False otherwise
    """
    log = logging.getLogger()

    r = requests.get(
        f"{RAILS_API_ENDPOINT}/api_user",
        headers={"Authorization": f"Token {api_key}"},
    )

    if not r.ok:
        log.error(
            "Error in verifying API Key.\nReason: %s, Message: %s", r.reason, r.text
        )

    return r.ok



from aicrowd.utils.utils import (
    is_subsequence,
    AliasedGroup,
    RichProgressBar,
    TqdmProgressBar,
)

__all__ = [
    "is_subsequence",
    "AliasedGroup",
    "RichProgressBar",
    "TqdmProgressBar",
]



"""
This class describes the config for the CLI
"""

import logging
import os
from pathlib import Path

import click
import toml

from aicrowd.errors import INVALID_FILE, CONFIG_NOT_SET
from aicrowd.exceptions import CLIException


class CLIConfig:
    """
    Stores all the configuration items
    """

    def __init__(self):
        self.log: logging.Logger = logging.getLogger()
        self.config_path: str = None
        self.__settings: dict = None

    def load(self, config_path: str):
        """
        Loads the config file

        Args:
            config_path: path to config file
        """
        if config_path is None:
            self.config_path = os.path.join(
                click.get_app_dir("aicrowd-cli"), "config.toml"
            )
            Path(self.config_path).parent.mkdir(parents=True, exist_ok=True)
            Path(self.config_path).touch()
        elif os.path.isfile(config_path):
            self.config_path = config_path
        else:
            self.log.error("Specified config file, '%s' doesn't exist", config_path)
            raise CLIException(
                f"Config file '{config_path}' doesn't exist", exit_code=INVALID_FILE
            )

        self.log.info("Loading config from %s", self.config_path)

        try:
            with open(self.config_path, "r") as config:
                self.__settings = toml.load(config)
                self.log.info("Config loaded\n---\n%s\n---", self.__settings)
        except toml.TomlDecodeError:
            self.log.error(
                "Malformed config file: %s. Will be overwritten", self.config_path
            )
            self.__settings = {}

    def write(self):
        """
        Save config to file
        """
        # make sure parent dir exists
        basedir = Path(self.config_path).parent.absolute()
        Path(basedir).mkdir(parents=True, exist_ok=True)

        with open(self.config_path, "w+") as config_file:
            toml.dump(self.__settings, config_file)
            self.log.info(
                "Successfully saved config to %s\n---\n%s\n---",
                self.config_path,
                self.__settings,
            )

    def get(self, key: str, ensure_exists: bool = False, help_msg: str = "") -> str:
        """
        Gets a key from config

        Prints help_msg if ensure_exists is True and key not found

        Args:
            key: the config setting you are looking for
            ensure_exists: should this die if key doesn't exist?
            help_msg: how to fix this issue

        Returns:
            the value for the key
        """
        if self.__settings is None:
            self.load(self.config_path)

        value = self.__settings.get(key)

        if value is None and ensure_exists:
            self.log.warning("Queried config for %s but not found", key)
            raise CLIException(
                f"Config property {key} not set", help_msg, CONFIG_NOT_SET
            )

        return value

    def set(self, key: str, value: str):
        """
        Sets a key to value in config

        Args:
            key: config key
            value: config value
        """
        if self.__settings is None:
            self.load(self.config_path)

        self.__settings[key] = value
